@@ -18,7 +18,9 @@ import (
 
 	"github.com/prometheus/prometheus/discovery/azure"
 	"github.com/prometheus/prometheus/discovery/consul"
+	"github.com/prometheus/prometheus/discovery/digitalocean"
 	"github.com/prometheus/prometheus/discovery/dns"
+	"github.com/prometheus/prometheus/discovery/dockerswarm"
 	"github.com/prometheus/prometheus/discovery/ec2"
 	"github.com/prometheus/prometheus/discovery/file"
 	"github.com/prometheus/prometheus/discovery/gce"
@@ -40,6 +42,10 @@ type ServiceDiscoveryConfig struct {
 	FileSDConfigs []*file.SDConfig `yaml:"file_sd_configs,omitempty"`
 	// List of Consul service discovery configurations.
 	ConsulSDConfigs []*consul.SDConfig `yaml:"consul_sd_configs,omitempty"`
+	// List of DigitalOcean service discovery configurations.
+	DigitalOceanSDConfigs []*digitalocean.SDConfig `yaml:"digitalocean_sd_configs,omitempty"`
+	// List of Docker Swarm service discovery configurations.
+	DockerSwarmSDConfigs []*dockerswarm.SDConfig `yaml:"dockerswarm_sd_configs,omitempty"`
 	// List of Serverset service discovery configurations.
 	ServersetSDConfigs []*zookeeper.ServersetSDConfig `yaml:"serverset_sd_configs,omitempty"`
 	// NerveSDConfigs is a list of Nerve service discovery configurations.
@@ -72,6 +78,16 @@ func (c *ServiceDiscoveryConfig) Validate() error {
 			return errors.New("empty or null section in consul_sd_configs")
 		}
 	}
+	for _, cfg := range c.DigitalOceanSDConfigs {
+		if cfg == nil {
+			return errors.New("empty or null section in digitalocean_sd_configs")
+		}
+	}
+	for _, cfg := range c.DockerSwarmSDConfigs {
+		if cfg == nil {
+			return errors.New("empty or null section in dockerswarm_sd_configs")
+		}
+	}
 	for _, cfg := range c.DNSSDConfigs {
 		if cfg == nil {
 			return errors.New("empty or null section in dns_sd_configs")
