@@ -20,12 +20,13 @@ import (
 	"time"
 
 	"github.com/go-kit/kit/log"
-	"github.com/prometheus/prometheus/discovery/targetgroup"
-	"github.com/prometheus/prometheus/util/testutil"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/client-go/kubernetes"
 	"k8s.io/client-go/kubernetes/fake"
 	"k8s.io/client-go/tools/cache"
+
+	"github.com/prometheus/prometheus/discovery/targetgroup"
+	"github.com/prometheus/prometheus/util/testutil"
 )
 
 // makeDiscovery creates a kubernetes.Discovery instance for testing.
@@ -65,6 +66,21 @@ func (d k8sDiscoveryTest) Run(t *testing.T) {
 
 	// Run discoverer and start a goroutine to read results.
 	go d.discovery.Run(ctx, ch)
+
+	// Ensure that discovery has a discoverer set. This prevents a race
+	// condition where the above go routine may or may not have set a
+	// discoverer yet.
+	for {
+		dis := d.discovery.(*Discovery)
+		dis.RLock()
+		l := len(dis.discoverers)
+		dis.RUnlock()
+		if l > 0 {
+			break
+		}
+		time.Sleep(10 * time.Millisecond)
+	}
+
 	resChan := make(chan map[string]*targetgroup.Group)
 	go readResultWithTimeout(t, ch, d.expectedMaxItems, time.Second, resChan)
 
@@ -91,14 +107,18 @@ func (d k8sDiscoveryTest) Run(t *testing.T) {
 // readResultWithTimeout reads all targegroups from channel with timeout.
 // It merges targegroups by source and sends the result to result channel.
 func readResultWithTimeout(t *testing.T, ch <-chan []*targetgroup.Group, max int, timeout time.Duration, resChan chan<- map[string]*targetgroup.Group) {
-	allTgs := make([][]*targetgroup.Group, 0)
-
+	res := make(map[string]*targetgroup.Group)
 Loop:
 	for {
 		select {
 		case tgs := <-ch:
-			allTgs = append(allTgs, tgs)
-			if len(allTgs) == max {
+			for _, tg := range tgs {
+				if tg == nil {
+					continue
+				}
+				res[tg.Source] = tg
+			}
+			if len(res) == max {
 				// Reached max target groups we may get, break fast.
 				break Loop
 			}
@@ -106,21 +126,11 @@ Loop:
 			// Because we use queue, an object that is created then
 			// deleted or updated may be processed only once.
 			// So possibly we may skip events, timed out here.
-			t.Logf("timed out, got %d (max: %d) items, some events are skipped", len(allTgs), max)
+			t.Logf("timed out, got %d (max: %d) items, some events are skipped", len(res), max)
 			break Loop
 		}
 	}
 
-	// Merge by source and sent it to channel.
-	res := make(map[string]*targetgroup.Group)
-	for _, tgs := range allTgs {
-		for _, tg := range tgs {
-			if tg == nil {
-				continue
-			}
-			res[tg.Source] = tg
-		}
-	}
 	resChan <- res
 }
 
