@@ -29,7 +29,9 @@ import (
 
 	"github.com/prometheus/prometheus/discovery/azure"
 	"github.com/prometheus/prometheus/discovery/consul"
+	"github.com/prometheus/prometheus/discovery/digitalocean"
 	"github.com/prometheus/prometheus/discovery/dns"
+	"github.com/prometheus/prometheus/discovery/dockerswarm"
 	"github.com/prometheus/prometheus/discovery/ec2"
 	"github.com/prometheus/prometheus/discovery/file"
 	"github.com/prometheus/prometheus/discovery/gce"
@@ -41,10 +43,10 @@ import (
 )
 
 var (
-	failedConfigs = prometheus.NewCounterVec(
-		prometheus.CounterOpts{
-			Name: "prometheus_sd_configs_failed_total",
-			Help: "Total number of service discovery configurations that failed to load.",
+	failedConfigs = prometheus.NewGaugeVec(
+		prometheus.GaugeOpts{
+			Name: "prometheus_sd_failed_configs",
+			Help: "Current number of service discovery configurations that failed to load.",
 		},
 		[]string{"name"},
 	)
@@ -191,10 +193,17 @@ func (m *Manager) ApplyConfig(cfg map[string]sd_config.ServiceDiscoveryConfig) e
 		}
 	}
 	m.cancelDiscoverers()
+	m.targets = make(map[poolKey]map[string]*targetgroup.Group)
+	m.providers = nil
+	m.discoverCancel = nil
+
+	failedCount := 0
 	for name, scfg := range cfg {
-		m.registerProviders(scfg, name)
+		failedCount += m.registerProviders(scfg, name)
 		discoveredTargets.WithLabelValues(m.name, name).Set(0)
 	}
+	failedConfigs.WithLabelValues(m.name).Set(float64(failedCount))
+
 	for _, prov := range m.providers {
 		m.startProvider(m.ctx, prov)
 	}
@@ -232,7 +241,7 @@ func (m *Manager) updater(ctx context.Context, p *provider, updates chan []*targ
 		case tgs, ok := <-updates:
 			receivedUpdates.WithLabelValues(m.name).Inc()
 			if !ok {
-				level.Debug(m.logger).Log("msg", "discoverer channel closed", "provider", p.name)
+				level.Debug(m.logger).Log("msg", "Discoverer channel closed", "provider", p.name)
 				return
 			}
 
@@ -264,7 +273,7 @@ func (m *Manager) sender() {
 				case m.syncCh <- m.allGroups():
 				default:
 					delayedUpdates.WithLabelValues(m.name).Inc()
-					level.Debug(m.logger).Log("msg", "discovery receiver's channel was full so will retry the next cycle")
+					level.Debug(m.logger).Log("msg", "Discovery receiver's channel was full so will retry the next cycle")
 					select {
 					case m.triggerSend <- struct{}{}:
 					default:
@@ -280,28 +289,25 @@ func (m *Manager) cancelDiscoverers() {
 	for _, c := range m.discoverCancel {
 		c()
 	}
-	m.targets = make(map[poolKey]map[string]*targetgroup.Group)
-	m.providers = nil
-	m.discoverCancel = nil
 }
 
 func (m *Manager) updateGroup(poolKey poolKey, tgs []*targetgroup.Group) {
 	m.mtx.Lock()
 	defer m.mtx.Unlock()
 
+	if _, ok := m.targets[poolKey]; !ok {
+		m.targets[poolKey] = make(map[string]*targetgroup.Group)
+	}
 	for _, tg := range tgs {
 		if tg != nil { // Some Discoverers send nil target group so need to check for it to avoid panics.
-			if _, ok := m.targets[poolKey]; !ok {
-				m.targets[poolKey] = make(map[string]*targetgroup.Group)
-			}
 			m.targets[poolKey][tg.Source] = tg
 		}
 	}
 }
 
 func (m *Manager) allGroups() map[string][]*targetgroup.Group {
-	m.mtx.Lock()
-	defer m.mtx.Unlock()
+	m.mtx.RLock()
+	defer m.mtx.RUnlock()
 
 	tSets := map[string][]*targetgroup.Group{}
 	for pkey, tsets := range m.targets {
@@ -317,8 +323,12 @@ func (m *Manager) allGroups() map[string][]*targetgroup.Group {
 	return tSets
 }
 
-func (m *Manager) registerProviders(cfg sd_config.ServiceDiscoveryConfig, setName string) {
-	var added bool
+// registerProviders returns a number of failed SD config.
+func (m *Manager) registerProviders(cfg sd_config.ServiceDiscoveryConfig, setName string) int {
+	var (
+		failedCount int
+		added       bool
+	)
 	add := func(cfg interface{}, newDiscoverer func() (Discoverer, error)) {
 		t := reflect.TypeOf(cfg).String()
 		for _, p := range m.providers {
@@ -332,7 +342,7 @@ func (m *Manager) registerProviders(cfg sd_config.ServiceDiscoveryConfig, setNam
 		d, err := newDiscoverer()
 		if err != nil {
 			level.Error(m.logger).Log("msg", "Cannot create service discovery", "err", err, "type", t)
-			failedConfigs.WithLabelValues(m.name).Inc()
+			failedCount++
 			return
 		}
 
@@ -361,6 +371,16 @@ func (m *Manager) registerProviders(cfg sd_config.ServiceDiscoveryConfig, setNam
 			return consul.NewDiscovery(c, log.With(m.logger, "discovery", "consul"))
 		})
 	}
+	for _, c := range cfg.DigitalOceanSDConfigs {
+		add(c, func() (Discoverer, error) {
+			return digitalocean.NewDiscovery(c, log.With(m.logger, "discovery", "digitalocean"))
+		})
+	}
+	for _, c := range cfg.DockerSwarmSDConfigs {
+		add(c, func() (Discoverer, error) {
+			return dockerswarm.NewDiscovery(c, log.With(m.logger, "discovery", "dockerswarm"))
+		})
+	}
 	for _, c := range cfg.MarathonSDConfigs {
 		add(c, func() (Discoverer, error) {
 			return marathon.NewDiscovery(*c, log.With(m.logger, "discovery", "marathon"))
@@ -421,6 +441,7 @@ func (m *Manager) registerProviders(cfg sd_config.ServiceDiscoveryConfig, setNam
 			return &StaticProvider{TargetGroups: []*targetgroup.Group{{}}}, nil
 		})
 	}
+	return failedCount
 }
 
 // StaticProvider holds a list of target groups that never change.
