@@ -17,6 +17,7 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
+	"io"
 	"io/ioutil"
 	"net/http"
 	"net/url"
@@ -24,12 +25,12 @@ import (
 	"time"
 
 	"github.com/go-kit/kit/log"
-	"github.com/go-kit/kit/log/level"
-	"github.com/mwitkow/go-conntrack"
-	"github.com/prometheus/client_golang/prometheus"
+	conntrack "github.com/mwitkow/go-conntrack"
+	"github.com/pkg/errors"
+	config_util "github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
 
-	config_util "github.com/prometheus/common/config"
+	"github.com/prometheus/prometheus/discovery/refresh"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 )
 
@@ -43,28 +44,18 @@ const (
 	tritonLabelServerID     = tritonLabel + "server_id"
 )
 
-var (
-	refreshFailuresCount = prometheus.NewCounter(
-		prometheus.CounterOpts{
-			Name: "prometheus_sd_triton_refresh_failures_total",
-			Help: "The number of Triton-SD scrape failures.",
-		})
-	refreshDuration = prometheus.NewSummary(
-		prometheus.SummaryOpts{
-			Name: "prometheus_sd_triton_refresh_duration_seconds",
-			Help: "The duration of a Triton-SD refresh in seconds.",
-		})
-	// DefaultSDConfig is the default Triton SD configuration.
-	DefaultSDConfig = SDConfig{
-		Port:            9163,
-		RefreshInterval: model.Duration(60 * time.Second),
-		Version:         1,
-	}
-)
+// DefaultSDConfig is the default Triton SD configuration.
+var DefaultSDConfig = SDConfig{
+	Role:            "container",
+	Port:            9163,
+	RefreshInterval: model.Duration(60 * time.Second),
+	Version:         1,
+}
 
 // SDConfig is the configuration for Triton based service discovery.
 type SDConfig struct {
 	Account         string                `yaml:"account"`
+	Role            string                `yaml:"role"`
 	DNSSuffix       string                `yaml:"dns_suffix"`
 	Endpoint        string                `yaml:"endpoint"`
 	Groups          []string              `yaml:"groups,omitempty"`
@@ -82,26 +73,24 @@ func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	if err != nil {
 		return err
 	}
+	if c.Role != "container" && c.Role != "cn" {
+		return errors.New("triton SD configuration requires role to be 'container' or 'cn'")
+	}
 	if c.Account == "" {
-		return fmt.Errorf("triton SD configuration requires an account")
+		return errors.New("triton SD configuration requires an account")
 	}
 	if c.DNSSuffix == "" {
-		return fmt.Errorf("triton SD configuration requires a dns_suffix")
+		return errors.New("triton SD configuration requires a dns_suffix")
 	}
 	if c.Endpoint == "" {
-		return fmt.Errorf("triton SD configuration requires an endpoint")
+		return errors.New("triton SD configuration requires an endpoint")
 	}
 	if c.RefreshInterval <= 0 {
-		return fmt.Errorf("triton SD configuration requires RefreshInterval to be a positive integer")
+		return errors.New("triton SD configuration requires RefreshInterval to be a positive integer")
 	}
 	return nil
 }
 
-func init() {
-	prometheus.MustRegister(refreshFailuresCount)
-	prometheus.MustRegister(refreshDuration)
-}
-
 // DiscoveryResponse models a JSON response from the Triton discovery.
 type DiscoveryResponse struct {
 	Containers []struct {
@@ -114,21 +103,25 @@ type DiscoveryResponse struct {
 	} `json:"containers"`
 }
 
+// ComputeNodeDiscoveryResponse models a JSON response from the Triton discovery /gz/ endpoint.
+type ComputeNodeDiscoveryResponse struct {
+	ComputeNodes []struct {
+		ServerUUID     string `json:"server_uuid"`
+		ServerHostname string `json:"server_hostname"`
+	} `json:"cns"`
+}
+
 // Discovery periodically performs Triton-SD requests. It implements
 // the Discoverer interface.
 type Discovery struct {
+	*refresh.Discovery
 	client   *http.Client
 	interval time.Duration
-	logger   log.Logger
 	sdConfig *SDConfig
 }
 
 // New returns a new Discovery which periodically refreshes its targets.
 func New(logger log.Logger, conf *SDConfig) (*Discovery, error) {
-	if logger == nil {
-		logger = log.NewNopLogger()
-	}
-
 	tls, err := config_util.NewTLSConfig(&conf.TLSConfig)
 	if err != nil {
 		return nil, err
@@ -143,79 +136,88 @@ func New(logger log.Logger, conf *SDConfig) (*Discovery, error) {
 	}
 	client := &http.Client{Transport: transport}
 
-	return &Discovery{
+	d := &Discovery{
 		client:   client,
 		interval: time.Duration(conf.RefreshInterval),
-		logger:   logger,
 		sdConfig: conf,
-	}, nil
+	}
+	d.Discovery = refresh.NewDiscovery(
+		logger,
+		"triton",
+		time.Duration(conf.RefreshInterval),
+		d.refresh,
+	)
+	return d, nil
 }
 
-// Run implements the Discoverer interface.
-func (d *Discovery) Run(ctx context.Context, ch chan<- []*targetgroup.Group) {
-	defer close(ch)
+// triton-cmon has two discovery endpoints:
+// https://github.com/joyent/triton-cmon/blob/master/lib/endpoints/discover.js
+//
+// The default endpoint exposes "containers", otherwise called "virtual machines" in triton,
+// which are (branded) zones running on the triton platform.
+//
+// The /gz/ endpoint exposes "compute nodes", also known as "servers" or "global zones",
+// on which the "containers" are running.
+//
+// As triton is not internally consistent in using these names,
+// the terms as used in triton-cmon are used here.
 
-	ticker := time.NewTicker(d.interval)
-	defer ticker.Stop()
-
-	// Get an initial set right away.
-	tg, err := d.refresh()
-	if err != nil {
-		level.Error(d.logger).Log("msg", "Refreshing targets failed", "err", err)
-	} else {
-		ch <- []*targetgroup.Group{tg}
+func (d *Discovery) refresh(ctx context.Context) ([]*targetgroup.Group, error) {
+	var endpointFormat string
+	switch d.sdConfig.Role {
+	case "container":
+		endpointFormat = "https://%s:%d/v%d/discover"
+	case "cn":
+		endpointFormat = "https://%s:%d/v%d/gz/discover"
+	default:
+		return nil, errors.New(fmt.Sprintf("unknown role '%s' in configuration", d.sdConfig.Role))
 	}
-
-	for {
-		select {
-		case <-ticker.C:
-			tg, err := d.refresh()
-			if err != nil {
-				level.Error(d.logger).Log("msg", "Refreshing targets failed", "err", err)
-			} else {
-				ch <- []*targetgroup.Group{tg}
-			}
-		case <-ctx.Done():
-			return
-		}
-	}
-}
-
-func (d *Discovery) refresh() (tg *targetgroup.Group, err error) {
-	t0 := time.Now()
-	defer func() {
-		refreshDuration.Observe(time.Since(t0).Seconds())
-		if err != nil {
-			refreshFailuresCount.Inc()
-		}
-	}()
-
-	var endpoint = fmt.Sprintf("https://%s:%d/v%d/discover", d.sdConfig.Endpoint, d.sdConfig.Port, d.sdConfig.Version)
+	var endpoint = fmt.Sprintf(endpointFormat, d.sdConfig.Endpoint, d.sdConfig.Port, d.sdConfig.Version)
 	if len(d.sdConfig.Groups) > 0 {
 		groups := url.QueryEscape(strings.Join(d.sdConfig.Groups, ","))
 		endpoint = fmt.Sprintf("%s?groups=%s", endpoint, groups)
 	}
 
-	tg = &targetgroup.Group{
-		Source: endpoint,
-	}
-
-	resp, err := d.client.Get(endpoint)
+	req, err := http.NewRequest("GET", endpoint, nil)
 	if err != nil {
-		return tg, fmt.Errorf("an error occurred when requesting targets from the discovery endpoint. %s", err)
+		return nil, err
+	}
+	req = req.WithContext(ctx)
+	resp, err := d.client.Do(req)
+	if err != nil {
+		return nil, errors.Wrap(err, "an error occurred when requesting targets from the discovery endpoint")
 	}
 
-	defer resp.Body.Close()
+	defer func() {
+		io.Copy(ioutil.Discard, resp.Body)
+		resp.Body.Close()
+	}()
 
 	data, err := ioutil.ReadAll(resp.Body)
 	if err != nil {
-		return tg, fmt.Errorf("an error occurred when reading the response body. %s", err)
+		return nil, errors.Wrap(err, "an error occurred when reading the response body")
+	}
+
+	// The JSON response body is different so it needs to be processed/mapped separately.
+	switch d.sdConfig.Role {
+	case "container":
+		return d.processContainerResponse(data, endpoint)
+	case "cn":
+		return d.processComputeNodeResponse(data, endpoint)
+	default:
+		return nil, errors.New(fmt.Sprintf("unknown role '%s' in configuration", d.sdConfig.Role))
+	}
+}
+
+func (d *Discovery) processContainerResponse(data []byte, endpoint string) ([]*targetgroup.Group, error) {
+	tg := &targetgroup.Group{
+		Source: endpoint,
 	}
 
 	dr := DiscoveryResponse{}
-	err = json.Unmarshal(data, &dr)
+	err := json.Unmarshal(data, &dr)
 	if err != nil {
-		return tg, fmt.Errorf("an error occurred unmarshaling the disovery response json. %s", err)
+		return nil, errors.Wrap(err, "an error occurred unmarshaling the discovery response json")
 	}
 
 	for _, container := range dr.Containers {
@@ -237,5 +239,30 @@ func (d *Discovery) refresh() (tg *targetgroup.Group, err error) {
 		tg.Targets = append(tg.Targets, labels)
 	}
 
-	return tg, nil
+	return []*targetgroup.Group{tg}, nil
+}
+
+func (d *Discovery) processComputeNodeResponse(data []byte, endpoint string) ([]*targetgroup.Group, error) {
+	tg := &targetgroup.Group{
+		Source: endpoint,
+	}
+
+	dr := ComputeNodeDiscoveryResponse{}
+	err := json.Unmarshal(data, &dr)
+	if err != nil {
+		return nil, errors.Wrap(err, "an error occurred unmarshaling the compute node discovery response json")
+	}
+
+	for _, cn := range dr.ComputeNodes {
+		labels := model.LabelSet{
+			tritonLabelMachineID:    model.LabelValue(cn.ServerUUID),
+			tritonLabelMachineAlias: model.LabelValue(cn.ServerHostname),
+		}
+		addr := fmt.Sprintf("%s.%s:%d", cn.ServerUUID, d.sdConfig.DNSSuffix, d.sdConfig.Port)
+		labels[model.AddressLabel] = model.LabelValue(addr)
+
+		tg.Targets = append(tg.Targets, labels)
+	}
+
+	return []*targetgroup.Group{tg}, nil
 }
